
/*
	OneStroke -- osGTK.c
	Copyright 2007 Michael Levin <risujin@risujin.org>

	NOTE: This project has been succeeded by CellWriter, a grid-entry
	      handwriting input panel. Please see:
	        http://risujin.org/cellwriter
*/

/*
    This file is part of OneStroke.

    OneStroke is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OneStroke is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OneStroke; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* Standard library headers */
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

/* X11 headers */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/extensions/XTest.h>

/* GDK/GTK headers */
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>

/* OneStroke header */
#include "onestroke.h"


/*
	Key Codes
*/

/* Modifier key codes */
static KeyCode kc_shift,kc_ctrl,kc_alt,kc_win,kc_super,kc_hyper;

/* Return a string representation of a symbol's key code */
char* osSymKeyString(osSymbol* sym) {
	char *text = 0,*key;
	text = malloc(64);
	key = XKeysymToString(sym->key);
	snprintf(text,64,"%s%s%s%s%s",
			sym->flags&OS_SHIFT?"Shift+":"",
			sym->flags&OS_CTRL?"Ctrl+":"",
			sym->flags&OS_ALT?"Alt+":"",
			sym->flags&OS_WIN?"Win+":"",
			key?key:"(none)");
	return text;
}


/*
	Fatal Error
*/

void _fatal(char* file,int line,char *format,...) {
	char buf[1024];
	va_list va;
	va_start(va,format);
	snprintf(buf,1023,"onestroke (%s:%d): %s\n",file,line,format);
	vfprintf(stderr,buf,va);
	va_end(va);
	exit(1);
}


/*
	Main Window
*/

static GtkWidget *window,*menu;

static void destroy(GtkWidget *widget,gpointer data) {
	gtk_main_quit();
}

/* Show about dialog */
static void showAbout(void) {
	char *authors[] = {"Michael Levin\n<cyren@mn.rr.com>\n"
			"http://www.missirina.com/risujin/contact.php\n",NULL},
		*artists[] = {"Eugenia Loli (OneStroke icon)\n"
			"http://eugenia.blogsome.com/",NULL};
	gtk_show_about_dialog(GTK_WINDOW(window),
		"license",
    "OneStroke -- a freehand gesture character input system.\n"
    "Copyright (C) 2006  Michael Levin\n\n"
    "This program is free software; you can redistribute it and/or modify"
    "it under the terms of the GNU General Public License as published by"
    "the Free Software Foundation; either version 2 of the License, or"
    "(at your option) any later version.\n\n"
    "This program is distributed in the hope that it will be useful,"
    "but WITHOUT ANY WARRANTY; without even the implied warranty of"
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the"
    "GNU General Public License for more details.\n\n"
    "You should have received a copy of the GNU General Public License along"
    "with this program; if not, write to the Free Software Foundation, Inc.,"
    "51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.",
    	"wrap-license",TRUE,
		"version",OS_VERSION,
		"website","http://www.missirina.com/risujin/onestroke.php ",
		"name","OneStroke",
		"logo-icon-name","onestroke",
		"copyright","Copyright (C) 2006 Michael Levin",
		"authors",authors,
		"artists",artists,
		"comments","A freehand gesture character input system.",
 		NULL);
}


/*
	Symbols Window
*/

#define ICON_WIDTH	48
#define ICON_HEIGHT	48
#define ICON_BPS	8
#define ICON_BRIGHT	224

static GdkPixmap *pixmap;
static GtkWidget *drawing_area,*sym_window,*sym_icons,*sym_combo,*sym_mod[4],
	*sym_scroll,*sym_vbox,*sym_delete,*sym_profiles,*sym_save_p,*sym_load_p,
	*sym_delete_p;
static GtkListStore *sym_icon_store,*sym_combo_store;
static GtkTreeIter sym_icon_iter;
static osSymbol *sym_selected = 0;

/* Icon view columns */
enum {
	SIS_TEXT,
	SIS_PIXBUF,
	SIS_SYMBOL,
	SIS_COLUMNS
};

/* Combo box columns */
enum {
	SCS_TEXT,
	SCS_KEY,
	SCS_COLUMNS
};

/* Set various effects of modification/file existance */
static void setModified(void) {

	/* Symbols window title */
	if(os_modified) gtk_window_set_title(GTK_WINDOW(sym_window),
		"Symbols - OneStroke " OS_VERSION "*");
	else gtk_window_set_title(GTK_WINDOW(sym_window),
		"Symbols - OneStroke " OS_VERSION);

	/* Symbols window profile buttons */
	gtk_widget_set_sensitive(sym_save_p,os_modified?TRUE:FALSE);
	gtk_widget_set_sensitive(sym_load_p,os_exists?TRUE:FALSE);
	gtk_widget_set_sensitive(sym_delete_p,os_exists?TRUE:FALSE);
}

/* Clear the drawing area */
static void clearDrawingArea(void) {
	gdk_draw_rectangle(pixmap,drawing_area->style->white_gc,TRUE,0,0,
		drawing_area->allocation.width,drawing_area->allocation.height);
	gtk_widget_queue_draw_area(drawing_area,0,0,
		drawing_area->allocation.width,drawing_area->allocation.height);
}

/* Draw a symbol showing point spacing */
static void drawPoints(osSymbol *symbol) {
	static GdkGC *gc;
	GdkColor color;
	int i;
	if(!gc) gc = gdk_gc_new(pixmap);

	/* Draw lines/points */
	for(i=0;i<symbol->length-1;i++) {
		int x1=symbol->points[i].x-OS_SCALE/2+os_match_cx+1,
			y1=symbol->points[i].y-OS_SCALE/2+os_match_cy+1,
			x2=symbol->points[i+1].x-OS_SCALE/2+os_match_cx+1,
			y2=symbol->points[i+1].y-OS_SCALE/2+os_match_cy+1;
		color.blue = 32767;
		if(i<=symbol->length/2) {
			color.red = 32767+65535*i/symbol->length;
			color.green = 65535;
		} else {
			color.red = 65535;
			color.green = 32767+65535*(symbol->length-i)/symbol->length;
		}
		gdk_gc_set_rgb_fg_color(gc,&color);
		gdk_draw_line(pixmap,gc,x1,y1,x2,y2);
		gdk_draw_point(pixmap,drawing_area->style->black_gc,x1,y1);
	}
	gdk_draw_point(pixmap,drawing_area->style->black_gc,
		symbol->points[symbol->length-1].x-OS_SCALE/2+os_match_cx+1,
		symbol->points[symbol->length-1].y-OS_SCALE/2+os_match_cy+1);

	/* Queue dirty area */
	gtk_widget_queue_draw_area(drawing_area,0,0,
		drawing_area->allocation.width,drawing_area->allocation.height);
}

/* Draw a symbol showing angles */
static void drawAngles(osSymbol *symbol) {
	static GdkGC *gc;
	GdkColor color;
	int i;
	if(!gc) gc = gdk_gc_new(pixmap);

	/* Draw lines */
	for(i=0;i<symbol->length-1;i++) {
		int x1 = (int)symbol->points[i].x-OS_SCALE/2+os_match_cx+1,
			y1 = (int)symbol->points[i].y-OS_SCALE/2+os_match_cy+1,
			x2 = (int)symbol->points[i+1].x-OS_SCALE/2+os_match_cx+1,
			y2 = (int)symbol->points[i+1].y-OS_SCALE/2+os_match_cy+1;
		float prop=symbol->segments[i].angle/M_PI*3.f;
		if(prop<0.f) prop += 6.;
		if(prop<1.f) {
			color.red = 65535;
			color.green = (unsigned short)(65535.f*prop);
			color.blue = 0;
		} else if(prop<2.f) {
			color.red = (unsigned short)(65535.f*(2.f-prop));
			color.green = 65535;
			color.blue = 0;
		} else if(prop<3.f) {
			color.red = 0;
			color.green = 65535;
			color.blue = (unsigned short)(65535.f*(prop-2.f));
		} else if(prop<4.f) {
			color.red = 0;
			color.green = (unsigned short)(65535.f*(4.f-prop));
			color.blue = 65535;
		} else if(prop<5.f) {
			color.red = (unsigned short)(65535.f*(prop-4.f));
			color.green = 0;
			color.blue = 65535;
		} else {
			color.red = 65535;
			color.green = 0;
			color.blue = (unsigned short)(65535.f*(6.f-prop));
		}
		gdk_gc_set_rgb_fg_color(gc,&color);
		gdk_draw_line(pixmap,gc,x1,y1,x2,y2);
	}

	/* Queue dirty area */
	gtk_widget_queue_draw_area(drawing_area,0,0,
		drawing_area->allocation.width,drawing_area->allocation.height);
}

/* Draw a symbol using the current mode */
void drawSymbol(osSymbol *symbol) {
	if(os_flags) {
		if(os_flags==1) drawAngles(symbol);
		else if(os_flags==2) drawPoints(symbol);
	}
}

/* Clear drawing area and draw a symbol centered */
void drawSymCenter(osSymbol *symbol) {
	clearDrawingArea();
	os_match_cx = drawing_area->allocation.width/2;
	os_match_cy = drawing_area->allocation.height/2;
	drawSymbol(symbol);
}

/* Open the symbol setup window */
static void sym_showWindow(void) {
	debug("Opening symbol window...");
	setModified();
	gtk_widget_show(sym_window);
}

/* Hide the symmbol setup window */
static gboolean sym_hideWindow(GtkWidget* widget) {
	gtk_widget_hide(widget);
	sym_selected = 0;
	debug("Symbol window closed.");
	return TRUE;
}

/* Add a new symbol */
static void sym_newSymbol(void) {
	GtkTreeIter iter;
	GtkTreePath *path;
	osSymbol *sym = osSymNew();
	osSymsetAdd(sym);
	gtk_list_store_append(sym_icon_store,&iter);
	gtk_list_store_set(sym_icon_store,&iter,SIS_TEXT,"(none)",
		SIS_PIXBUF,0,SIS_SYMBOL,sym,-1);
	path = gtk_tree_model_get_path(GTK_TREE_MODEL(sym_icon_store),&iter);
	gtk_icon_view_select_path(GTK_ICON_VIEW(sym_icons),path);
	gtk_tree_path_free(path);
	setModified();
}

/* Delete selected symbol */
static void sym_deleteSymbol(void) {
	if(sym_selected) {
		osSymFree(sym_selected);
		gtk_list_store_remove(sym_icon_store,&sym_icon_iter);
		clearDrawingArea();
		setModified();
	}
}

/* Update the text of the selected icon */
static void sym_updateText(GtkTreeIter *iter) {
	GValue gv_symbol;
	memset(&gv_symbol,0,sizeof(gv_symbol));
	gtk_tree_model_get_value(GTK_TREE_MODEL(sym_icon_store),
		iter,SIS_SYMBOL,&gv_symbol);
	gtk_list_store_set(sym_icon_store,iter,SIS_TEXT,
		osSymKeyString(g_value_get_pointer(&gv_symbol)),-1);
}

/* A modifier has been toggled */
static void sym_modToggle(GtkToggleButton *button,int mod) {
	if(sym_selected) {
		gboolean active = gtk_toggle_button_get_active(button);
		if(active && !(sym_selected->flags&mod)) {
			osSymKeyChange(sym_selected,sym_selected->key,
				sym_selected->flags|mod);
			sym_updateText(&sym_icon_iter);
		} else if(!active && sym_selected->flags&mod) {
			osSymKeyChange(sym_selected,sym_selected->key,
				sym_selected->flags&(~mod));
			sym_updateText(&sym_icon_iter);
		}
		setModified();
	}
}

/* The icon view selection has changed */
static void sym_iconSelect(GtkIconView *iconview) {
	GList *list = gtk_icon_view_get_selected_items(iconview);
	if(list) {
		GValue value;
		GtkTreePath *path = g_list_nth_data(list,0);
		GtkTreeIter iter;

		/* Get selected symbol */
		memset(&value,0,sizeof(value));
		gtk_tree_model_get_iter(GTK_TREE_MODEL(sym_icon_store),
			&sym_icon_iter,path);
		gtk_tree_model_get_value(GTK_TREE_MODEL(sym_icon_store),
			&sym_icon_iter,SIS_SYMBOL,&value);
		sym_selected = g_value_get_pointer(&value);
		g_list_foreach(list,(GFunc)gtk_tree_path_free,NULL);
		g_list_free(list);

		/* Draw symbol */
		drawSymCenter(sym_selected);

		/* Setup combo box */
		gtk_combo_box_set_active(GTK_COMBO_BOX(sym_combo),-1);
		if(sym_selected->key && gtk_tree_model_get_iter_first(
				GTK_TREE_MODEL(sym_combo_store),&iter)) {
			int key;
			do {
				memset(&value,0,sizeof(value));
				gtk_tree_model_get_value(GTK_TREE_MODEL(sym_combo_store),
					&iter,SCS_KEY,&value);
				key = g_value_get_int(&value);
				if(sym_selected->key==key) {
					gtk_combo_box_set_active_iter(GTK_COMBO_BOX(sym_combo),
						&iter);
					break;
				}
			} while(gtk_tree_model_iter_next(
				GTK_TREE_MODEL(sym_combo_store),&iter));
		}

		/* Setup toggles accordingly */
		gtk_widget_set_sensitive(sym_vbox,TRUE);
		gtk_widget_set_sensitive(sym_delete,TRUE);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sym_mod[0]),
			(sym_selected->flags&OS_SHIFT)!=0);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sym_mod[1]),
			(sym_selected->flags&OS_CTRL)!=0);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sym_mod[2]),
			(sym_selected->flags&OS_ALT)!=0);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sym_mod[3]),
			(sym_selected->flags&OS_WIN)!=0);
	} else {
		sym_selected = 0;
		gtk_widget_set_sensitive(sym_vbox,FALSE);
		gtk_widget_set_sensitive(sym_delete,FALSE);
	}
}

/* Render the symbol pixbuf */
static void sym_updatePixbuf(GtkTreeIter *iter) {
	GValue gv_symbol,gv_pixbuf;
	GdkPixbuf *pixbuf;
	osSymbol *symbol;

	/* Retrieve symbol and pixbuf values */
	memset(&gv_symbol,0,sizeof(gv_symbol));
	memset(&gv_pixbuf,0,sizeof(gv_pixbuf));
	gtk_tree_model_get_value(GTK_TREE_MODEL(sym_icon_store),
		iter,SIS_SYMBOL,&gv_symbol);
	gtk_tree_model_get_value(GTK_TREE_MODEL(sym_icon_store),
		iter,SIS_PIXBUF,&gv_pixbuf);
	symbol = g_value_get_pointer(&gv_symbol);
	pixbuf = g_value_get_object(&gv_pixbuf);
	if(pixbuf) g_object_unref(pixbuf);

	/* Create new pixbuf */
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,TRUE,8,ICON_WIDTH,ICON_HEIGHT);
	if(pixbuf) {
		guchar *pixels = gdk_pixbuf_get_pixels(pixbuf);
		int i,j,l,row_stride = gdk_pixbuf_get_rowstride(pixbuf),
			channels = gdk_pixbuf_get_n_channels(pixbuf);

		/*debug("Drawing symbol with row stride %d, %d channels, %d points.",
			  row_stride,channels,symbol->length);*/

		/* Clear icon */
		memset(pixels,ICON_BRIGHT,ICON_WIDTH*channels);
		for(i=1;i<ICON_HEIGHT-1;i++) {
			guchar *p = pixels+i*row_stride;
			p[0] = ICON_BRIGHT;
			p[1] = ICON_BRIGHT;
			p[2] = ICON_BRIGHT;
			p[3] = ICON_BRIGHT;
			memset(p+channels,0,(ICON_WIDTH-2)*channels);
			p += row_stride-channels;
			p[0] = ICON_BRIGHT;
			p[1] = ICON_BRIGHT;
			p[2] = ICON_BRIGHT;
			p[3] = ICON_BRIGHT;
		}
		memset(pixels+(ICON_HEIGHT-1)*row_stride,ICON_BRIGHT,
			ICON_WIDTH*channels);

		/* Render line segments */
		for(j=0;j<symbol->length-1;j=l) {
			int points=0,k;

			/* Find the next endpoint and the number of points we're drawing */
			for(l=j+1;;l++) {
				int xpoints,ypoints,
					x1 = symbol->points[j].x,y1 = symbol->points[j].y,
					x2 = symbol->points[l].x,y2 = symbol->points[l].y;
				xpoints = x2*ICON_WIDTH/OS_SCALE-x1*ICON_WIDTH/OS_SCALE;
				ypoints = y2*ICON_HEIGHT/OS_SCALE-y1*ICON_HEIGHT/OS_SCALE;
				if(xpoints<0) xpoints = -xpoints;
				if(ypoints<0) ypoints = -ypoints;
				points = xpoints>ypoints?xpoints:ypoints;
				if(points>0||l>=symbol->length-1) break;
			}

			/* Draw the line, point by point */
			for(k=0;k<points;k++) {
				guchar *p;
				int x,y,
					x1 = symbol->points[j].x,y1 = symbol->points[j].y,
					x2 = symbol->points[l].x,y2 = symbol->points[l].y;
				x1 = x1*ICON_WIDTH/OS_SCALE;
				y1 = y1*ICON_HEIGHT/OS_SCALE;
				x2 = x2*ICON_WIDTH/OS_SCALE;
				y2 = y2*ICON_HEIGHT/OS_SCALE;
				x = (x1*(points-k)+x2*k)/points,
				y = (y1*(points-k)+y2*k)/points;
				p = pixels+y*row_stride+x*channels;
				if(j<=symbol->length/2) {
					p[0] = ICON_BRIGHT*2*j/symbol->length;
					p[1] = ICON_BRIGHT;
				} else {
					p[0] = ICON_BRIGHT;
					p[1] = ICON_BRIGHT*2*(symbol->length-j)/symbol->length;
				}
				p[3] = 255;
			}
		}

		gtk_list_store_set(sym_icon_store,iter,SIS_PIXBUF,pixbuf,-1);
	} else
		debug("Failed to create new pixbuf.");
}

/* Populate icon view with items */
static void sym_populate(void) {
	GtkTreeIter iter;
	gtk_list_store_clear(sym_icon_store);
	osSymbol *sym = osSymsetFirst();
	while(sym) {
		gtk_list_store_append(sym_icon_store,&iter);
		gtk_list_store_set(sym_icon_store,&iter,SIS_TEXT,
			osSymKeyString(sym),SIS_PIXBUF,0,SIS_SYMBOL,sym,-1);
		sym_updatePixbuf(&iter);
		sym = osSymsetNext(sym);
	}
}

/* Save profile */
static void sym_saveProfile(void) {
	if(!osProfileSave()&&
			gtk_combo_box_get_active(GTK_COMBO_BOX(sym_profiles))==-1) {
		gchar *text = gtk_combo_box_get_active_text(
			GTK_COMBO_BOX(sym_profiles));
		gtk_combo_box_prepend_text(GTK_COMBO_BOX(sym_profiles),text);
		gtk_combo_box_set_active(GTK_COMBO_BOX(sym_profiles),0);
	}
	setModified();
}

/* Delete profile */
static void sym_deleteProfile(void) {
	guint active = gtk_combo_box_get_active(GTK_COMBO_BOX(sym_profiles));
	if(active!=-1) {
		osProfileDelete();
		gtk_combo_box_remove_text(GTK_COMBO_BOX(sym_profiles),active);
		gtk_combo_box_set_active(GTK_COMBO_BOX(sym_profiles),-1);
		gtk_list_store_clear(sym_icon_store);
		clearDrawingArea();
		setModified();
	}
}

/* Set profile name */
static void sym_profChange(GtkComboBox *combo) {
	gchar *text = gtk_combo_box_get_active_text(combo);
	osProfileSet(text);
	setModified();
}

/* The key combo box selection has changed */
static void sym_comboSelect(GtkComboBox *combo) {
	GtkTreeIter iter;
	GValue gv_key;
	int key;
	memset(&gv_key,0,sizeof(gv_key));
	if(gtk_combo_box_get_active_iter(combo,&iter)) {
		gtk_tree_model_get_value(GTK_TREE_MODEL(sym_combo_store),
			&iter,SCS_KEY,&gv_key);
		key = g_value_get_int(&gv_key);
		if(key!=sym_selected->key) {
			osSymKeyChange(sym_selected,key,sym_selected->flags);
			sym_updateText(&sym_icon_iter);
			setModified();
		}
	}
}


/*
	Drawing Area Widget
*/

#define TIMEOUT_BOUNDS	1000
#define TIMEOUT_WAIT	500
#define TIMEOUT_REPEAT	100

static gboolean drawing = FALSE;
static guint cancel_event = 0,wait_event = 0,repeat_event = 0;
static int last_x,last_y;

/* Create a new backing pixmap of the appropriate size */
static gboolean configure_event(GtkWidget *widget,GdkEventConfigure *event) {
	if(pixmap) g_object_unref(pixmap);
	pixmap = gdk_pixmap_new(widget->window,widget->allocation.width,
		widget->allocation.height,-1);
	if(sym_selected) drawSymCenter(sym_selected);
	else clearDrawingArea();
	return TRUE;
}

/* Redraw the drawing area from the backing pixmap */
static gboolean expose_event(GtkWidget *widget,GdkEventExpose *event) {
	gdk_draw_drawable(widget->window,
		widget->style->fg_gc[GTK_WIDGET_STATE(widget)],pixmap,
		event->area.x,event->area.y,event->area.x,event->area.y,
		event->area.width,event->area.height);
	return FALSE;
}

/* Draw a segment of the stroke */
static void drawSegment(int x,int y) {
	osPoint pt;

	/* Correct coordinates */
	if(x<0) x = 0;
	else if(x>=drawing_area->allocation.width)
		x = drawing_area->allocation.width-1;
	if(y<0) y = 0;
	else if(y>=drawing_area->allocation.height)
		y = drawing_area->allocation.height-1;

	/* Add a node */
	pt.x = x;
	pt.y = y;
	/*debug("Drawing point %d,%d real %d,%d.",pt.x,pt.y,x,y);*/
	osCurDraw(pt);

	if(pixmap) {
		int x1,x2,y1,y2,xmin,xmax,ymin,ymax;
		x1 = last_x;
		y1 = last_y;
		x2 = x;
		y2 = y;

		/* Find minimum and maximum x */
		if(x1>x2) { xmax = x1; xmin = x2; }
		else { xmin = x1; xmax = x2; }

		/* Find minimum and maximum y */
		if(y1>y2) { ymax = y1; ymin = y2; }
		else { ymin = y1; ymax = y2; }

		/* Draw the new segment */
		gdk_draw_line(pixmap,drawing_area->style->black_gc,x1,y1,x2,y2);
		gtk_widget_queue_draw_area(drawing_area,xmin,ymin,
			xmax-xmin+1,ymax-ymin+1);
	}
	last_x = x;
	last_y = y;
}

/* Mouse button is pressed over drawing area */
static gboolean button_press_event(GtkWidget *widget,GdkEventButton *event) {
	if(!drawing && event->button==1) {
		osPoint pt;
		drawing = TRUE;
		last_x = event->x;
		last_y = event->y;
		pt.x = event->x;
		pt.y = event->y;
		clearDrawingArea();
		osCurDraw(pt);
	} else if(event->button==3) {
		debug("Opening context menu...");
		gtk_menu_popup(GTK_MENU(menu),0,0,0,0,event->button,event->time);
	}
	return TRUE;
}

/* Send a key event */
static gboolean sendEvent(osSymbol* symbol) {
	KeyCode kc = XKeysymToKeycode(GDK_DISPLAY(),symbol->key);

	/* Send key down events */
	if(symbol->flags&OS_SHIFT)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_shift,True,0);
	if(symbol->flags&OS_CTRL)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_ctrl,True,0);
	if(symbol->flags&OS_ALT)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_alt,True,0);
	if(symbol->flags&OS_WIN)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_win,True,0);
	if(symbol->flags&OS_SUPER)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_super,True,0);
	if(symbol->flags&OS_HYPER)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_hyper,True,0);
	XTestFakeKeyEvent(GDK_DISPLAY(),kc,True,0);

	/* Send key up events */
	XTestFakeKeyEvent(GDK_DISPLAY(),kc,False,0);
	if(symbol->flags&OS_SHIFT)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_shift,False,0);
	if(symbol->flags&OS_CTRL)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_ctrl,False,0);
	if(symbol->flags&OS_ALT)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_alt,False,0);
	if(symbol->flags&OS_WIN)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_win,False,0);
	if(symbol->flags&OS_SUPER)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_super,False,0);
	if(symbol->flags&OS_HYPER)
		XTestFakeKeyEvent(GDK_DISPLAY(),kc_hyper,False,0);

	return TRUE;
}

/* Cancel key repeat events */
static void cancelRepeat(void) {
	if(repeat_event) {
		g_source_remove(repeat_event);
		repeat_event = 0;
	}
	if(wait_event) {
		g_source_remove(wait_event);
		wait_event = 0;
	}
}

/* Cancel mouse stroke */
static gboolean cancelStroke(gpointer data) {
	osCurReset();
	clearDrawingArea();
	drawing = FALSE;
	cancel_event = 0;
	debug("Pen stroke canceled.");
	return FALSE;
}

/* Wait for repeat event */
static gboolean waitEvent(gpointer data) {
	if(!sym_selected) {
		osSymbol *sym = osCurMatch();
		if(sym) {
			repeat_event = g_timeout_add(TIMEOUT_REPEAT,
				(GSourceFunc)sendEvent,sym);
			drawing = FALSE;
			drawSymbol(sym);
		} else
			debug("Symbol unmatched.");
	}
	return FALSE;
}

/* Mouse button is released over drawing area */
static gboolean button_release_event(GtkWidget *widget,GdkEventButton *event) {
	if(drawing && event->button==1) {
		GtkTreeIter iter;
		drawing = FALSE;

		/* Clear the window */
		clearDrawingArea();

		/* Train symbol */
		if(sym_selected) {
			osSymTrain(sym_selected);
			sym_updatePixbuf(&sym_icon_iter);
			drawSymCenter(sym_selected);
			setModified();
		}

		/* Send key event */
		else {
			osSymbol *symbol = osCurMatch();
			if(symbol) {
				sendEvent(symbol);
				drawSymbol(symbol);
			}
			else debug("Stroke unrecognized.");

			/* Update stroke pixbuf */
			if(symbol) {
				gtk_tree_model_get_iter_first(GTK_TREE_MODEL(sym_icon_store),
					&iter);
				do {
					GValue gv_symbol;
					osSymbol *ps;
					memset(&gv_symbol,0,sizeof(gv_symbol));
					gtk_tree_model_get_value(GTK_TREE_MODEL(sym_icon_store),
						&iter,SIS_SYMBOL,&gv_symbol);
					ps = g_value_get_pointer(&gv_symbol);
					if(ps==symbol) {
						sym_updatePixbuf(&iter);
						break;
					}
				} while(gtk_tree_model_iter_next(GTK_TREE_MODEL(sym_icon_store),
					&iter));
			}
		}

		/* Draw the current stroke */
		drawSymbol(os_cur_stroke);
	}

	/* Reset */
	osCurReset();
	cancelRepeat();
	if(cancel_event) g_source_remove(cancel_event);

	return TRUE;
}

/* Mouse is moved over drawing area */
static gboolean motion_notify_event(GtkWidget *widget,GdkEventMotion *event) {
	gint x,y;
	GdkModifierType state;
	if(event->is_hint)
		gdk_window_get_pointer(event->window,&x,&y,&state);
	else {
		x = event->x;
		y = event->y;
		state = event->state;
	}
	if(drawing) {
		if(wait_event) {
			g_source_remove(wait_event);
			wait_event = 0;
		}
		if(!cancel_event&&!sym_selected)
			wait_event = g_timeout_add(TIMEOUT_WAIT,waitEvent,0);
		drawSegment(x,y);
	}
	return TRUE;
}

/* Mouse leaves the drawing area */
static gboolean leave_notify_event(GtkWidget *widget) {
	if(drawing) {
		cancelRepeat();
		cancel_event = g_timeout_add(TIMEOUT_BOUNDS,cancelStroke,0);
	}
	return TRUE;
}

/* Mouse enters the drawing area */
static gboolean enter_notify_event(GtkWidget *widget) {
	if(drawing&&cancel_event) {
		g_source_remove(cancel_event);
		debug("Pen stroke out of bounds canceled.");
	}
	return TRUE;
}


/*
	Frontend
*/

#define INIT_WIDTH	128
#define INIT_HEIGHT	128

static GtkWidget *menu_mode[OS_MODES],*menu_draw[3],*menu_rotate,*menu_favor;

static gboolean sym_keyPress(GtkWidget *widget,GdkEventKey *event) {
	switch(event->keyval) {
		case GDK_d:
		case GDK_D:
			osSymDump(sym_selected);
			break;
		case GDK_Delete:
			sym_deleteSymbol();
			break;
	}
	return FALSE;
}

void setDraw(GtkCheckMenuItem *check) {
	g_signal_handlers_block_by_func(menu_draw[0],setDraw,0);
	g_signal_handlers_block_by_func(menu_draw[1],setDraw,0);
	g_signal_handlers_block_by_func(menu_draw[2],setDraw,0);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_draw[0]),FALSE);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_draw[1]),FALSE);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_draw[2]),FALSE);
	if(check==GTK_CHECK_MENU_ITEM(menu_draw[0])) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_draw[0]),
			TRUE);
		os_flags = 0;
		debug("Not drawing.");
	} else if(check==GTK_CHECK_MENU_ITEM(menu_draw[1])) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_draw[1]),
			TRUE);
		os_flags = 1;
		debug("Drawing angles.");
	} else if(check==GTK_CHECK_MENU_ITEM(menu_draw[2])) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_draw[2]),
			TRUE);
		os_flags = 2;
		debug("Drawing points.");
	}
	if(sym_selected) drawSymCenter(sym_selected);
	g_signal_handlers_unblock_by_func(menu_draw[0],setDraw,0);
	g_signal_handlers_unblock_by_func(menu_draw[1],setDraw,0);
	g_signal_handlers_unblock_by_func(menu_draw[2],setDraw,0);
}

void setMode(GtkCheckMenuItem *check);
void setRotate(int on,int favor) {
	g_signal_handlers_block_by_func(menu_rotate,setMode,0);
	g_signal_handlers_block_by_func(menu_favor,setMode,0);
	if(on) {
		gtk_check_menu_item_set_active(
			GTK_CHECK_MENU_ITEM(menu_rotate),TRUE);
		gtk_widget_set_sensitive(menu_favor,TRUE);
		os_mode |= OS_ROTATE;
		if(favor) {
			gtk_check_menu_item_set_active(
				GTK_CHECK_MENU_ITEM(menu_favor),TRUE);
			os_mode |= OS_FAVOR;
			debug("Rotational invariance on, favoring original angle.");
		} else {
			gtk_check_menu_item_set_active(
				GTK_CHECK_MENU_ITEM(menu_favor),FALSE);
			os_mode &= ~OS_FAVOR;
			debug("Rotational invariance on.");
		}
	} else {
		gtk_check_menu_item_set_active(
			GTK_CHECK_MENU_ITEM(menu_rotate),FALSE);
		gtk_widget_set_sensitive(menu_favor,FALSE);
		os_mode &= ~OS_ROTATE;
		debug("Rotational invariance off.");
	}
	g_signal_handlers_unblock_by_func(menu_rotate,setMode,0);
	g_signal_handlers_unblock_by_func(menu_favor,setMode,0);
}

void setMode(GtkCheckMenuItem *check) {
	int i;
	if(check) {

		/* Toggling rotational invariance */
		if(check==GTK_CHECK_MENU_ITEM(menu_rotate))
			setRotate(!(os_mode&OS_ROTATE),os_mode&OS_FAVOR);
		else if(check==GTK_CHECK_MENU_ITEM(menu_favor))
			setRotate(os_mode&OS_ROTATE,!(os_mode&OS_FAVOR));

		/* Selecting a mode */
		else {
			for(i=0;i<OS_MODES;i++) {
				g_signal_handlers_block_by_func(menu_mode[i],setMode,0);
				gtk_check_menu_item_set_active(
					GTK_CHECK_MENU_ITEM(menu_mode[i]),FALSE);
			}
			gtk_check_menu_item_set_active(check,TRUE);
			if(check==GTK_CHECK_MENU_ITEM(menu_mode[OS_TRAP_DIST])) {
				os_mode = OS_TRAP_DIST|(os_mode&OS_TOGGLE_MASK);
				debug("Trapezoid distance mode.");
			} else if(check==GTK_CHECK_MENU_ITEM(menu_mode[OS_TRAP_ANGLE])) {
				os_mode = OS_TRAP_ANGLE|(os_mode&OS_TOGGLE_MASK);
				debug("Trapezoid angle mode.");
			} else if(check==GTK_CHECK_MENU_ITEM(menu_mode[OS_TRAP_COMP])) {
				os_mode = OS_TRAP_COMP|(os_mode&OS_TOGGLE_MASK);
				debug("Trapezoid composite mode.");
			}
			for(i=0;i<OS_MODES;i++)
				g_signal_handlers_unblock_by_func(menu_mode[i],setMode,0);
		}

		os_modified = 1;
		setModified();
	}
}

/* Load profile */
static void sym_loadProfile(void) {
	if(!osProfileLoad()) {
		sym_populate();
		setDraw(GTK_CHECK_MENU_ITEM(menu_draw[os_flags]));
		setMode(GTK_CHECK_MENU_ITEM(menu_mode[os_mode&OS_MODE_MASK]));
		setRotate(os_mode&OS_ROTATE,os_mode&OS_FAVOR);
		os_modified = 0;
		setModified();
	}
}

void osFrontend(int argc,char *argv[]) {

	/* Initialize GTK */
	gtk_init (&argc, &argv);

	/* Create main window */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT(window),"destroy",G_CALLBACK(destroy),NULL);
	gtk_container_set_border_width(GTK_CONTAINER(window),2);
	gtk_window_resize(GTK_WINDOW(window),INIT_WIDTH,INIT_HEIGHT);
	gtk_widget_set_events(window,GDK_KEY_PRESS_MASK);
	gtk_window_set_type_hint(GTK_WINDOW(window),GDK_WINDOW_TYPE_HINT_UTILITY);
	gtk_window_set_keep_above(GTK_WINDOW(window),TRUE);
	gtk_window_set_accept_focus(GTK_WINDOW(window),FALSE);
	gtk_window_set_decorated(GTK_WINDOW(window),TRUE);
	gtk_window_set_title(GTK_WINDOW(window),"OneStroke");
	gtk_window_set_skip_pager_hint(GTK_WINDOW(window),TRUE);
	gtk_window_set_skip_taskbar_hint(GTK_WINDOW(window),TRUE);
	gtk_window_stick(GTK_WINDOW(window));
	gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_MOUSE);

	/* Window geometry hints */
	{
		GdkGeometry hints;
		hints.max_width = OS_SCALE;
		hints.max_height = OS_SCALE;
		gtk_window_set_geometry_hints(GTK_WINDOW(window),0,&hints,
			GDK_HINT_MAX_SIZE);
	}

	/* Create drawing area */
	drawing_area = gtk_drawing_area_new();
	g_signal_connect(G_OBJECT(drawing_area),"expose_event",
		G_CALLBACK(expose_event),NULL);
	g_signal_connect(G_OBJECT(drawing_area),"configure_event",
		G_CALLBACK(configure_event),NULL);
	g_signal_connect(G_OBJECT(drawing_area),"button_press_event",
		G_CALLBACK(button_press_event),NULL);
	g_signal_connect(G_OBJECT(drawing_area),"button_release_event",
		G_CALLBACK(button_release_event),NULL);
	g_signal_connect(G_OBJECT(drawing_area),"motion_notify_event",
		G_CALLBACK(motion_notify_event),NULL);
	g_signal_connect(G_OBJECT(drawing_area),"leave_notify_event",
		G_CALLBACK(leave_notify_event),NULL);
	g_signal_connect(G_OBJECT(drawing_area),"enter_notify_event",
		G_CALLBACK(enter_notify_event),NULL);
	gtk_widget_set_events(drawing_area,GDK_EXPOSURE_MASK
			|GDK_LEAVE_NOTIFY_MASK
			|GDK_BUTTON_PRESS_MASK
			|GDK_BUTTON_RELEASE_MASK
			|GDK_POINTER_MOTION_MASK
			|GDK_POINTER_MOTION_HINT_MASK);
	gtk_widget_set_extension_events(drawing_area,GDK_EXTENSION_EVENTS_CURSOR);
	gtk_container_add(GTK_CONTAINER(window),drawing_area);

	/* Create back-buffer pixmap */
	pixmap = gdk_pixmap_new(0,INIT_WIDTH,INIT_HEIGHT,24);

	/* Create symbol window */
	{
		GtkWidget *vbox,*hboxt,*vboxl,*hbox,*hboxl,*button,*label;

		/* Setup window */
		sym_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_transient_for(GTK_WINDOW(sym_window),GTK_WINDOW(window));
		gtk_window_resize(GTK_WINDOW(sym_window),512,320);
		gtk_container_set_border_width(GTK_CONTAINER(sym_window),0);
		gtk_container_set_border_width(GTK_CONTAINER(sym_window),4);
		g_signal_connect(G_OBJECT(sym_window),"delete_event",
			G_CALLBACK(sym_hideWindow),0);
		g_signal_connect(G_OBJECT(sym_window),"key_press_event",
			G_CALLBACK(sym_keyPress),0);

		/* Setup boxes */
		vbox = gtk_vbox_new(FALSE,4);
		hboxt = gtk_hbox_new(FALSE,4);
		hbox = gtk_hbox_new(FALSE,4);
		hboxl = gtk_hbox_new(FALSE,4);
		vboxl = gtk_vbox_new(FALSE,4);
		sym_vbox = gtk_vbox_new(FALSE,4);
		gtk_box_pack_start(GTK_BOX(hbox),vboxl,TRUE,TRUE,0);
		gtk_box_pack_start(GTK_BOX(hbox),sym_vbox,FALSE,FALSE,0);
		gtk_box_pack_end(GTK_BOX(vboxl),hboxl,FALSE,FALSE,0);
		gtk_box_pack_start(GTK_BOX(vbox),hboxt,FALSE,FALSE,0);
		gtk_box_pack_start(GTK_BOX(vbox),hbox,TRUE,TRUE,0);
		gtk_container_add(GTK_CONTAINER(sym_window),vbox);
		gtk_widget_show(vbox);
		gtk_widget_show(hboxt);
		gtk_widget_show(hbox);
		gtk_widget_show(hboxl);
		gtk_widget_show(vboxl);
		gtk_widget_show(sym_vbox);
		gtk_widget_set_sensitive(sym_vbox,FALSE);

		/* Setup profile label */
		label = gtk_label_new("Profile:");
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(hboxt),label,FALSE,FALSE,0);

		/* Profile delete button */
		sym_delete_p = gtk_button_new_with_label("Delete");
		g_signal_connect(G_OBJECT(sym_delete_p),"clicked",
			G_CALLBACK(sym_deleteProfile),NULL);
		gtk_widget_show(sym_delete_p);
		gtk_box_pack_end(GTK_BOX(hboxt),sym_delete_p,FALSE,FALSE,0);

		/* Profile save button */
		sym_save_p = gtk_button_new_with_label("Save");
		g_signal_connect(G_OBJECT(sym_save_p),"clicked",
			G_CALLBACK(sym_saveProfile),NULL);
		gtk_widget_show(sym_save_p);
		gtk_box_pack_end(GTK_BOX(hboxt),sym_save_p,FALSE,FALSE,0);

		/* Profile load button */
		sym_load_p = gtk_button_new_with_label("Load");
		g_signal_connect(G_OBJECT(sym_load_p),"clicked",
			G_CALLBACK(sym_loadProfile),NULL);
		gtk_widget_show(sym_load_p);
		gtk_box_pack_end(GTK_BOX(hboxt),sym_load_p,FALSE,FALSE,0);

		/* Setup and fill profile combo box */
		{
			char *text;
			sym_profiles = gtk_combo_box_entry_new_text();
			while((text=osGetProfile()))
				gtk_combo_box_append_text(GTK_COMBO_BOX(sym_profiles),text);
			gtk_widget_show(sym_profiles);
			gtk_box_pack_start(GTK_BOX(hboxt),sym_profiles,TRUE,TRUE,0);
			g_signal_connect(G_OBJECT(sym_profiles),"changed",
				G_CALLBACK(sym_profChange),NULL);
			gtk_combo_box_set_active(GTK_COMBO_BOX(sym_profiles),0);
		}

		/* Setup icon view */
		sym_icon_store = gtk_list_store_new(SIS_COLUMNS,
			G_TYPE_STRING,GDK_TYPE_PIXBUF,G_TYPE_POINTER);
		gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(sym_icon_store),
			SIS_TEXT,GTK_SORT_ASCENDING);
		sym_icons = gtk_icon_view_new_with_model(GTK_TREE_MODEL(
			sym_icon_store));
		g_signal_connect(G_OBJECT(sym_icons),"selection-changed",
			G_CALLBACK(sym_iconSelect),NULL);
		gtk_icon_view_set_selection_mode(GTK_ICON_VIEW(sym_icons),
			GTK_SELECTION_SINGLE);
		gtk_icon_view_set_text_column(GTK_ICON_VIEW(sym_icons),SIS_TEXT);
		gtk_icon_view_set_pixbuf_column(GTK_ICON_VIEW(sym_icons),
			SIS_PIXBUF);
		gtk_widget_show(sym_icons);

		/* Setup scrolled window for icon view */
		sym_scroll = gtk_scrolled_window_new(NULL,NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sym_scroll),
			GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);
		gtk_box_pack_start(GTK_BOX(vboxl),sym_scroll,TRUE,TRUE,0);
		gtk_container_add(GTK_CONTAINER(sym_scroll),sym_icons);
		gtk_widget_show(sym_scroll);

		/* Setup combo label */
		label = gtk_label_new("Keycode to send:");
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(sym_vbox),label,FALSE,FALSE,0);

		/* Setup key combo box */
		{
			GtkCellRenderer *text_renderer = gtk_cell_renderer_text_new();
			sym_combo_store = gtk_list_store_new(SCS_COLUMNS,G_TYPE_STRING,
				G_TYPE_INT);
			gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(
				sym_combo_store),SCS_TEXT,GTK_SORT_ASCENDING);
			sym_combo = gtk_combo_box_new_with_model(
				GTK_TREE_MODEL(sym_combo_store));
			gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(sym_combo),
				text_renderer,TRUE);
			gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(sym_combo),
				text_renderer,"text",SCS_TEXT,(void*)0);
			g_signal_connect(G_OBJECT(sym_combo),"changed",
				G_CALLBACK(sym_comboSelect),NULL);
			gtk_widget_show(sym_combo);
			gtk_box_pack_start(GTK_BOX(sym_vbox),sym_combo,FALSE,TRUE,0);
		}

		/* Fill key combo box with key entries */
		{
			GtkTreeIter iter;
			KeySym *keysyms;
			int i,keysyms_per_keycode,min_keycode,max_keycode;
			XDisplayKeycodes(GDK_DISPLAY(),&min_keycode,&max_keycode);
			keysyms = XGetKeyboardMapping(GDK_DISPLAY(),min_keycode,
				max_keycode-min_keycode+1,&keysyms_per_keycode);
			for(i=min_keycode;i<max_keycode;i++) {
				KeySym ks = keysyms[(i-min_keycode)*keysyms_per_keycode];
				char *text = XKeysymToString(ks);
				if(text) {
					gtk_list_store_append(sym_combo_store,&iter);
					gtk_list_store_set(sym_combo_store,&iter,SCS_TEXT,
						text,SCS_KEY,ks,-1);
				}
			}
			XFree(keysyms);
		}

		/* Setup checkboxes */
		sym_mod[0] = gtk_check_button_new_with_label("Shift");
		sym_mod[1] = gtk_check_button_new_with_label("Control");
		sym_mod[2] = gtk_check_button_new_with_label("Alt");
		sym_mod[3] = gtk_check_button_new_with_label("Win-Key");
		g_signal_connect(G_OBJECT(sym_mod[0]),"toggled",
			G_CALLBACK(sym_modToggle),(void*)OS_SHIFT);
		g_signal_connect(G_OBJECT(sym_mod[1]),"toggled",
			G_CALLBACK(sym_modToggle),(void*)OS_CTRL);
		g_signal_connect(G_OBJECT(sym_mod[2]),"toggled",
			G_CALLBACK(sym_modToggle),(void*)OS_ALT);
		g_signal_connect(G_OBJECT(sym_mod[3]),"toggled",
			G_CALLBACK(sym_modToggle),(void*)OS_WIN);
		gtk_widget_show(sym_mod[0]);
		gtk_widget_show(sym_mod[1]);
		gtk_widget_show(sym_mod[2]);
		gtk_widget_show(sym_mod[3]);
		gtk_box_pack_start(GTK_BOX(sym_vbox),sym_mod[0],FALSE,TRUE,0);
		gtk_box_pack_start(GTK_BOX(sym_vbox),sym_mod[1],FALSE,TRUE,0);
		gtk_box_pack_start(GTK_BOX(sym_vbox),sym_mod[2],FALSE,TRUE,0);
		gtk_box_pack_start(GTK_BOX(sym_vbox),sym_mod[3],FALSE,TRUE,0);

		/* Info label */
		label = gtk_label_new("Draw on the main\nwindow to record\n"
			"a new gesture.");
		gtk_label_set_line_wrap(GTK_LABEL(label),TRUE);
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(sym_vbox),label,FALSE,TRUE,0);

		/* Symbols label */
		label = gtk_label_new("Symbol:");
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(hboxl),label,FALSE,FALSE,0);

		/* Setup add/delete buttons */
		button = gtk_button_new_with_label("New");
		sym_delete = gtk_button_new_with_label("Delete");
		g_signal_connect(G_OBJECT(button),"clicked",
			G_CALLBACK(sym_newSymbol),NULL);
		g_signal_connect(G_OBJECT(sym_delete),"clicked",
			G_CALLBACK(sym_deleteSymbol),NULL);
		gtk_box_pack_start(GTK_BOX(hboxl),button,TRUE,TRUE,0);
		gtk_box_pack_start(GTK_BOX(hboxl),sym_delete,TRUE,TRUE,0);
		gtk_widget_show(button);
		gtk_widget_show(sym_delete);
		gtk_widget_set_sensitive(sym_delete,FALSE);
	}

	/* Create context menu */
	{
		GtkWidget *item,*submenu;

		/* Symbols... */
		menu = gtk_menu_new();
		item = gtk_menu_item_new_with_label("Symbols...");
		g_signal_connect(G_OBJECT(item),"activate",
			G_CALLBACK(sym_showWindow),NULL);
		gtk_menu_attach(GTK_MENU(menu),item,0,1,0,1);

		/* Separator */
		item = gtk_separator_menu_item_new();
		gtk_menu_attach(GTK_MENU(menu),item,0,1,1,2);

		/* Submenu for Draw -> */
		submenu = gtk_menu_new();

		/* Draw -> None */
		menu_draw[0] = gtk_check_menu_item_new_with_label("None");
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_draw[0]),
			TRUE);
		g_signal_connect(G_OBJECT(menu_draw[0]),"activate",
			G_CALLBACK(setDraw),0);
		gtk_check_menu_item_set_draw_as_radio(
			GTK_CHECK_MENU_ITEM(menu_draw[0]),TRUE);
		gtk_menu_attach(GTK_MENU(submenu),menu_draw[0],0,1,0,1);

		/* Draw -> Angles */
		menu_draw[1] = gtk_check_menu_item_new_with_label("Angles");
		g_signal_connect(G_OBJECT(menu_draw[1]),"activate",
			G_CALLBACK(setDraw),0);
		gtk_check_menu_item_set_draw_as_radio(
			GTK_CHECK_MENU_ITEM(menu_draw[1]),TRUE);
		gtk_menu_attach(GTK_MENU(submenu),menu_draw[1],0,1,1,2);

		/* Draw -> Points */
		menu_draw[2] = gtk_check_menu_item_new_with_label("Points");
		g_signal_connect(G_OBJECT(menu_draw[2]),"activate",
			G_CALLBACK(setDraw),0);
		gtk_check_menu_item_set_draw_as_radio(
			GTK_CHECK_MENU_ITEM(menu_draw[2]),TRUE);
		gtk_menu_attach(GTK_MENU(submenu),menu_draw[2],0,1,2,3);

		/* Draw -> */
		item = gtk_menu_item_new_with_label("Draw");
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(item),submenu);
		gtk_menu_attach(GTK_MENU(menu),item,0,1,2,3);

		/* Submenu for Mode -> */
		submenu = gtk_menu_new();

		/* Mode -> Trapezoid Composite */
		menu_mode[OS_TRAP_COMP] =
			gtk_check_menu_item_new_with_label(
				"Composite (default)");
		g_signal_connect(G_OBJECT(menu_mode[OS_TRAP_COMP]),"activate",
			G_CALLBACK(setMode),0);
		gtk_check_menu_item_set_draw_as_radio(
			GTK_CHECK_MENU_ITEM(menu_mode[OS_TRAP_COMP]),TRUE);
		gtk_menu_attach(GTK_MENU(submenu),menu_mode[OS_TRAP_COMP],0,1,0,1);

		/* Mode -> Trapezoid Distance */
		menu_mode[OS_TRAP_DIST] =
			gtk_check_menu_item_new_with_label(
				"Scale dependent");
		g_signal_connect(G_OBJECT(menu_mode[OS_TRAP_DIST]),"activate",
			G_CALLBACK(setMode),0);
		gtk_check_menu_item_set_draw_as_radio(
			GTK_CHECK_MENU_ITEM(menu_mode[OS_TRAP_DIST]),TRUE);
		gtk_menu_attach(GTK_MENU(submenu),menu_mode[OS_TRAP_DIST],0,1,1,2);

		/* Mode -> Trapezoid Angle */
		menu_mode[OS_TRAP_ANGLE] = gtk_check_menu_item_new_with_label(
			"Scale independent");
		g_signal_connect(G_OBJECT(menu_mode[OS_TRAP_ANGLE]),"activate",
			G_CALLBACK(setMode),0);
		gtk_check_menu_item_set_draw_as_radio(
			GTK_CHECK_MENU_ITEM(menu_mode[OS_TRAP_ANGLE]),TRUE);
		gtk_menu_attach(GTK_MENU(submenu),menu_mode[OS_TRAP_ANGLE],0,1,2,3);

		/* Check current mode */
		gtk_check_menu_item_set_active(
			GTK_CHECK_MENU_ITEM(menu_mode[os_mode&OS_MODE_MASK]),TRUE);

		/* Mode -> Separator */
		item = gtk_separator_menu_item_new();
		gtk_menu_attach(GTK_MENU(submenu),item,0,1,3,4);

		/* Mode -> Rotational Invariance */
		menu_rotate =
			gtk_check_menu_item_new_with_label("Rotational invariance");
		g_signal_connect(G_OBJECT(menu_rotate),"activate",
			G_CALLBACK(setMode),0);
		gtk_menu_attach(GTK_MENU(submenu),menu_rotate,0,1,4,5);

		/* Mode -> Favor original orientation */
		menu_favor =
			gtk_check_menu_item_new_with_label("Favor original orientation");
		g_signal_connect(G_OBJECT(menu_favor),"activate",
			G_CALLBACK(setMode),0);
		gtk_menu_attach(GTK_MENU(submenu),menu_favor,0,1,5,6);

		/* Check rotational invariance */
		setRotate(os_mode&OS_ROTATE,os_mode&OS_FAVOR);

		/* Mode -> */
		item = gtk_menu_item_new_with_label("Mode");
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(item),submenu);
		gtk_menu_attach(GTK_MENU(menu),item,0,1,3,4);

		/* Quit */
		item = gtk_menu_item_new_with_label("About...");
		g_signal_connect(G_OBJECT(item),"activate",
			G_CALLBACK(showAbout),NULL);
		gtk_menu_attach(GTK_MENU(menu),item,0,1,4,5);

		/* Separator */
		item = gtk_separator_menu_item_new();
		gtk_menu_attach(GTK_MENU(menu),item,0,1,5,6);

		/* Quit */
		item = gtk_menu_item_new_with_label("Quit");
		g_signal_connect(G_OBJECT(item),"activate",
			G_CALLBACK(destroy),NULL);
		gtk_menu_attach(GTK_MENU(menu),item,0,1,7,8);

		/* Show all menu widgets */
		gtk_widget_show_all(menu);
	}

	/* Query XTest extension */
	{
		int a,b,c,d;
		if(!XTestQueryExtension(GDK_DISPLAY(),&a,&b,&c,&d))
			fatal("XTest extension not supported.");
		else
			debug("XTest extension (%d.%d.%d.%d) supported.",a,b,c,d);
	}

	/* Fetch modifier keys */
	kc_shift = XKeysymToKeycode(GDK_DISPLAY(),XK_Shift_L);
	kc_ctrl = XKeysymToKeycode(GDK_DISPLAY(),XK_Control_L);
	kc_alt = XKeysymToKeycode(GDK_DISPLAY(),XK_Alt_L);
	kc_win = XKeysymToKeycode(GDK_DISPLAY(),XK_Meta_L);
	kc_super = XKeysymToKeycode(GDK_DISPLAY(),XK_Super_L);
	kc_hyper = XKeysymToKeycode(GDK_DISPLAY(),XK_Hyper_L);

	/* Load profile */
	sym_loadProfile();

	/* Run GTK GUI */
	gtk_widget_show_all(window);
	gtk_main();
}
