
/*
	OneStroke -- osUnix.c
	Copyright 2007 Michael Levin <risujin@risujin.org>

	NOTE: This project has been succeeded by CellWriter, a grid-entry
	      handwriting input panel. Please see:
	        http://risujin.org/cellwriter
*/

/*
    This file is part of OneStroke.

    OneStroke is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OneStroke is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OneStroke; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* Standard library headers */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* Unix headers */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>

/* OneStroke header */
#include "onestroke.h"


/*
	Debug
*/

#ifdef _DEBUG
void _assert(char* file,int line,char* text) {
	_debug(file,line,"Assertion '%s' failed!",text);
	raise(SIGILL);
}
#endif


/*
	Profile
*/

#define FILE_MAGIC		((int)('M'<<24)+('Y'<<16)+('S'<<8)+'O')
#define FILE_VERSION	3
#define ROOT_LEN		64
#define ROOT_POSTFIX	"/.onestroke-0.8"
#define PROFILE_LEN		32

int os_flags=0,os_modified=0,os_exists=0;
static char root[ROOT_LEN];
static char profile[PROFILE_LEN] = "";

/* Profile file header */
typedef struct {
	int magic,version,symbols,mode,flags;
} osFileHeader;

/* Symbol header */
typedef struct {
	double distance;
	int length,key;
	short flags;
	char strokes;
} osSymHeader;

/* Copy a file */
static int copyFile(char *from,char *to) {
	FILE *f_from,*f_to;
	int c;
	f_from = fopen(from,"r");
	if(!f_from) {
		debug("CopyFile() failed to open '%s' for reading.",from);
		return 1;
	}
	f_to = fopen(to,"w");
	if(!f_to) {
		debug("CopyFile() failed to open '%s' for writing.",to);
		return 2;
	}
	while((c=fgetc(f_from))!=EOF) fputc(c,f_to);
	fclose(f_from);
	fclose(f_to);
	debug("Copied file '%s' to '%s'.",from,to);
	return 0;
}

/* Setup home directory */
int osSetCWD(void) {
	DIR *dir;
	char *cwd;

	/* Check if we are already in the home directory */
	cwd = getcwd(NULL,0);

	/* Not in home directory */
	if(strcmp(cwd,root)) {
		free(cwd);

		/* Construct program folder path */
		strncpy(root,getenv("HOME"),ROOT_LEN);
		strncat(root,ROOT_POSTFIX,ROOT_LEN-1);

		/* Ensure program directory exists */
		dir = opendir(root);
		if(!dir) {
			debug("Program directory '%s' does not exist.",root);
			if(mkdir(root,S_IRWXU|S_IRWXG|S_IRWXO)) {
				debug("Failed to create program root directory '%s'.",root);
				perrno();
				return 0;
			} else
				debug("Program root directory created.");
		} else {
			debug("Program directory '%s' exists.",root);
			closedir(dir);
		}

		/* Change to program directory */
		if(chdir(root)) {
			debug("Failed to change to program root directory.");
			perrno();
			return 0;
		} else
			debug("Changed to program root directory.");

	}

	/* Already in home directory */
	else {
		debug("Current directory already program root.");
		free(cwd);
	}

	/* Copy default profiles if directory is empty */
	if(!osGetProfile()) {
		struct dirent *de;
		dir = opendir(OS_PROFDIR);
		if(dir) {
			char buf[NAME_MAX] = OS_PROFDIR;
			size_t len = strlen(buf);
			buf[len++] = '/';
			while((de=readdir(dir))) {
				if(de->d_type==DT_REG) {
					strncpy(buf+len,de->d_name,NAME_MAX-len);
					buf[NAME_MAX-1] = 0;
					copyFile(buf,de->d_name);
				}
			}
		} else debug("Failed to open default profiles directory '%s'.",
			OS_PROFDIR);
	} else while(osGetProfile());

	return 1;
}

/* Return a profile name */
char* osGetProfile(void) {
	static DIR *dir;
	static struct dirent *de;
	if(!dir) dir = opendir(".");
	if(dir) {
		do {
			de = readdir(dir);
		} while(de&&de->d_type!=DT_REG);
		if(de) return de->d_name;
		closedir(dir);
		dir = 0;
	}
	return 0;
}

/* Set profile string */
void osProfileSet(char *path) {
	FILE *file;
	strncpy(profile,path,PROFILE_LEN-1);
	profile[PROFILE_LEN-1] = 0;
	os_modified = 1;
	if((file=fopen(path,"r"))) {
		fclose(file);
		os_exists = 1;
	} else os_exists = 0;
}

/* Delete a symbol set */
void osProfileDelete(void) {
	osSymsetFree();
	unlink(profile);
	os_modified = 0;
	os_exists = 0;
	debug("Deleted profile '%s'.",profile);
}

/* Load a symbol set */
int osProfileLoad(void) {
	FILE *file;
	int i;
	osSymsetFree();
	file = fopen(profile,"r");
	if(file) {
		osFileHeader fileh;

		/* Read file header */
		fread(&fileh,sizeof(fileh),1,file);
		if(fileh.magic!=FILE_MAGIC) {
			debug("File '%s' is not a valid profile.",profile);
			fclose(file);
			return 2;
		}
		if(fileh.version!=FILE_VERSION) {
			debug("Profile '%s' has bad version (%d, should be %d).",
				profile,fileh.version,FILE_VERSION);
			fclose(file);
			return 3;
		}
		os_flags = fileh.flags;
		os_mode = fileh.mode;

		/* Read symbol data */
		for(i=0;i<fileh.symbols;i++) {
			osSymbol *sym;
			osSymHeader symh;
			if(fread(&symh,sizeof(symh),1,file)<1) {
				debug("Profile '%s' ends prematurely, %d/%d symbols read.",
					profile,i,fileh.symbols);
				break;
			}
			if(symh.length>0) {
				size_t ptsread;
				sym = osSymNew();
				sym->distance = symh.distance;
				sym->length = symh.length;
				sym->key = symh.key;
				sym->flags = symh.flags;
				sym->points = malloc(symh.length*sizeof(osPoint));
				ptsread = fread(sym->points,sizeof(osPoint),symh.length,file);
				if(ptsread<symh.length) {
					debug("Profile '%s' ends prematurely, %d/%d symbols "
						" and %d/%d points read.",profile,i,fileh.symbols,
						ptsread,symh.length);
					osSymFree(sym);
				}
				osSymProcess(sym,0,0);
				osSymsetAdd(sym);
			}
		}

		fclose(file);
		os_modified = 0;
		os_exists = 1;
		debug("Read %d symbols from profile '%s'.",i,profile);
		return 0;
	} else
		debug("Failed to open '%s' for reading.",profile);
	return 1;
}

/* Save a symbol set */
int osProfileSave(void) {
	if(!osSymsetFirst()) {
		debug("No symbols to save.");
		return 0;
	}
	if(!os_modified) {
		debug("Profile is not modified, not saving.");
		return 0;
	}
	if(profile[0]) {
		FILE *file = fopen(profile,"w");
		if(file) {
			osFileHeader fileh;
			osSymbol *sym;
			int qty = 0;

			/* Write symbol data */
			fseek(file,sizeof(fileh),SEEK_SET);
			sym = osSymsetFirst();
			while(sym) {
				osSymHeader symh;
				symh.distance = sym->distance;
				symh.length = sym->length;
				symh.key = sym->key;
				symh.flags = sym->flags;
				fwrite(&symh,sizeof(symh),1,file);
				fwrite(sym->points,sizeof(osPoint),sym->length,file);
				qty++;
				sym = osSymsetNext(sym);
			}

			/* Write file header */
			fseek(file,0,SEEK_SET);
			fileh.magic = FILE_MAGIC;
			fileh.version = FILE_VERSION;
			fileh.symbols = qty;
			fileh.flags = os_flags;
			fileh.mode = os_mode;
			fwrite(&fileh,sizeof(fileh),1,file);

			fclose(file);
			os_modified = 0;
			os_exists = 1;
			debug("Saved %d symbols as profile '%s'.",qty,profile);
			return 0;
		} else
			debug("Failed to open '%s' for writing.",profile);
	}
	return 1;
}
